from functools import reduce
from typing import Any, Callable


def compose(*fs: Callable[..., Any]) -> Callable[..., Any]:
    def compose2(f: Callable[..., Any], g: Callable[..., Any]) -> Callable[..., Any]:
        return lambda *a, **kw: f(g(*a, **kw))

    return reduce(compose2, fs)



import sys


class Sentinel:
    def __init__(self, name: str):
        self.name = name

    def __repr__(self) -> str:
        return f"<{sys.intern(str(self.name)).rsplit('.', 1)[-1]}>"


NOID = Sentinel("NoId")



from .requests import (
    notification,
    notification_json,
    request,
    request_hex,
    request_json,
    request_json_hex,
    request_json_random,
    request_json_uuid,
    request_random,
    request_uuid,
)
from .responses import Ok, Error, parse, parse_json



from typing import Any, Dict, Iterable, List, Union, NamedTuple
import json

from .utils import compose

Deserialized = Union[Dict[str, Any], List[Dict[str, Any]]]


class Ok(NamedTuple):
    result: Any
    id: Any

    def __repr__(self) -> str:
        return f"Ok(result={self.result!r}, id={self.id!r})"


class Error(NamedTuple):
    code: int
    message: str
    data: Any
    id: Any

    def __repr__(self) -> str:
        return f"Error(code={self.code!r}, message={self.message!r}, data={self.data!r}, id={self.id!r})"


Response = Union[Ok, Error]


def to_result(response: Dict[str, Any]) -> Response:
    return (
        Ok(response["result"], response["id"])
        if "result" in response
        else Error(
            response["error"]["code"],
            response["error"]["message"],
            response["error"].get("data"),
            response["id"],
        )
    )


def parse(response: Deserialized) -> Union[Response, Iterable[Response]]:
    return (
        map(to_result, response) if isinstance(response, list) else to_result(response)
    )


parse_json = compose(parse, json.loads)



"""setup.py"""
from setuptools import setup

with open("README.md") as f:
    README = f.read()

setup(
    author="Beau Barker",
    author_email="beau@explodinglabs.com",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="Send JSON-RPC requests",
    include_package_data=True,
    install_requires=[],
    license="MIT",
    long_description=README,
    long_description_content_type="text/markdown",
    name="jsonrpcclient",
    # Be PEP 561 compliant
    # https://mypy.readthedocs.io/en/stable/installed_packages.html#making-pep-561-compatible-packages
    package_data={"jsonrpcclient": ["response-schema.json", "py.typed"]},
    zip_safe=False,
    packages=["jsonrpcclient"],
    url="https://github.com/explodinglabs/jsonrpcclient",
    version="4.0.2",
)



import json
from functools import partial
from typing import Any, Dict, Iterator, Tuple, Union

from . import id_generators
from .sentinels import NOID
from .utils import compose


def notification_pure(
    method: str, params: Union[Dict[str, Any], Tuple[Any, ...]]
) -> Dict[str, Any]:
    return {
        "jsonrpc": "2.0",
        "method": method,
        **({"params": params} if params else {}),
    }


def notification(
    method: str, params: Union[Dict[str, Any], Tuple[Any, ...], None] = None
) -> Dict[str, Any]:
    return notification_pure(method, params if params else ())


notification_json = compose(json.dumps, notification)


def request_pure(
    id_generator: Iterator[Any],
    method: str,
    params: Union[Dict[str, Any], Tuple[Any, ...]],
    id: Any,
) -> Dict[str, Any]:
    return {
        "jsonrpc": "2.0",
        "method": method,
        **(
            {"params": list(params) if isinstance(params, tuple) else params}
            if params
            else {}
        ),
        "id": id if id is not NOID else next(id_generator),
    }


def request_impure(
    id_generator: Iterator[Any],
    method: str,
    params: Union[Dict[str, Any], Tuple[Any, ...], None] = None,
    id: Any = NOID,
) -> Dict[str, Any]:
    return request_pure(
        id_generator or id_generators.decimal(), method, params or (), id
    )


request_natural = partial(request_impure, id_generators.decimal())
request_hex = partial(request_impure, id_generators.hexadecimal())
request_random = partial(request_impure, id_generators.random())
request_uuid = partial(request_impure, id_generators.uuid())
request = request_natural

request_json = compose(json.dumps, request_natural)
request_json_hex = compose(json.dumps, request_hex)
request_json_random = compose(json.dumps, request_random)
request_json_uuid = compose(json.dumps, request_uuid)
